unit _actor;
// HALF-LIFE MODEL VIEWER
// ---------------------------------------------------------------------------
// 3D ACTOR ANIMATION UNIT FOR TMT PASCAL
// version: 07.jun.2005 for DOS/WINDOWS
// by Georgy Moshkin
// e-mail  : tmtlib@narod.ru
// www : http://www.tmtlib.narod.ru/
// ---------------------------------------------------------------------------

interface
uses _strman,_math,polyx2d,memA,graph;

{R}
{Q}
// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
type TtextureCoord = record
u:single;
v:single;
end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
type TactorVertex = record
parentId : byte;
coord    : array[0..2] of single;
norm     : array[0..2] of single;
//u,v      : single;
end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
type TactorTriangle = record
textureId : byte;
vertexId  : array[0..2] of word;
coord    : array[0..2] of TtextureCoord;
end;


// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
type Tanimation=object
nodes       : word;
count       : word;
frames      : word;
rotation    : ^array[0..2] of single;
translation : ^array[0..2] of single;

fix         : boolean;
fps         : single;
procedure alloc;
procedure free;
procedure upload(fname:string);
end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
type Tactor=object
nodes      : word;
vertexes   : word;
triangles  : word;
texturesN  : word;
nodeParent : ^byte;
vertex     : ^TactorVertex;
localized  : ^TactorVertex;
triangle   : ^TactorTriangle;
rotation    : ^array[0..2] of single;
translation : ^array[0..2] of single;
texturelist: string;
material   : ^integer;

m_absolute      : ^matrix;  // joints
m_relative      : ^matrix;
r_absolute      : ^matrix;  // joints
r_relative      : ^matrix;

{a}prevTranslation : ^array[0..2] of single;

curtime     : single;
curframe    : word;
delta       : single;

qPrev       : ^quaternion;
qCur        : ^quaternion;
qCur1       : ^quaternion;
qCur2       : ^quaternion;
qFinal      : quaternion;

walk        : single;

posx        : single;
posy        : single;
posz        : single;

angle       : single;
angleps     : single;
teta        : single;
tetaps      : single;

procedure alloc;
procedure free;
procedure upload(txpath,fname:string);
procedure render;
procedure renderNT;
procedure localize;
procedure animate(var ani:Tanimation;dir:integer);
procedure animateBlend(var ani,ani2:Tanimation;percent:single;dir:integer);
end;

var elapsed     : double;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
type Tobject=object
vertexes   : word;
triangles  : word;
texturesN  : word;
vertex     : ^TactorVertex;
{a} localized : ^TactorVertex;
triangle   : ^TactorTriangle;
texturelist: string;
material   : ^integer;

procedure alloc;
procedure free;
procedure upload(txpath,fname:string);
procedure render;
procedure renderNT;
procedure localize;
procedure attach(act:Tactor;node:byte);
end;

procedure MyStartGraph;
procedure MyAddLights;
procedure MySetCam;
procedure MyBIObegin;
procedure MyBIOend;
procedure MyExitGraph;

implementation
{$i polyx3d}

var
zzz        : Obj3d; // temporary polyx 3d object for ONE TRIANGLE PRIMITIVE

procedure MyStartGraph;
begin
  StartGraph;
end;

procedure MyAddLights;
begin
//   AddLG(-10000000,-10000000,0, 255,255,100, 400,400,400, 0,XLGPermanent);
//   AddLG(10000000,1000000,0, 100,255,255, 400,400,400, 0,XLGPermanent);
//   AddLG(-100000,-100000,0, 255,255,100, 40,40,40, 0,XLGPermanent);
  // AddLG(100000,10000,0, 100,255,255, 40,40,40, 0,XLGPermanent);
     addlg(0,-10000,0, 75,75,75, 400,400,00,0, XLGAMBIENT);
     addlg(-10000,-1000,0, 215,215,215, 600,600,600,0, XLGPERMANENT);


end;

procedure MySetCam;
begin
 SetCam(1000,12000,12000,0,0,0);
 Cam_See(0,-7000,0);
end;

procedure MyBIObegin;
begin
           bio(bfclpage);
           bio(bfresetprim);
end;

procedure MyBIOend;
begin
           bio(bfqueryrender);
     outtextxy(10,210,'3D model viewer by Georgy Moshkin');
           ShowFrame;
end;

procedure MyExitGraph;
begin
     Exitgraph;
end;


// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tactor.alloc;
begin

//triangle primitive for Polyx 3d engine
     zzz.init;
     zzz.create(3,1); // 3 vertexes = 1 face
     lockobj(zzz);

     setvert(0,
             0,0,0,
             255,255,255,255,
             0,0);
     setvert(1,
             1000,0,0,
             255,255,255,255,
             1023,0);
     setvert(2,
             1000,1000,0,
             255,255,255,255,
             1023,1023);

     setface(0,
             0,1,2,-1,
             GetStatus(UtGuroudTex,BlendNone,TcMULRGB,0,0,0),255);


getmem(nodeParent,SizeOf(nodeParent^)*(nodes));
getmem(vertex,SizeOf(vertex^)*(vertexes));
{a} getmem(localized,SizeOf(localized^)*(vertexes));
getmem(triangle,SizeOf(triangle^)*(triangles));

getmem(rotation,SizeOf(rotation^)*(nodes));
getmem(translation,SizeOf(translation^)*(nodes));
{a} getmem(m_absolute,SizeOf(m_absolute^)*(nodes));
    getmem(m_relative,SizeOf(m_relative^)*(nodes));
{a} getmem(r_absolute,SizeOf(r_absolute^)*(nodes));
    getmem(r_relative,SizeOf(r_relative^)*(nodes));

getmem(material,SizeOf(material^)*(texturesN));

getmem(qPrev,SizeOf(qPrev^)*(nodes));
getmem(qCur,SizeOf(qCur^)*(nodes));

{b} getmem(qCur1,SizeOf(qCur1^)*(nodes));
{b} getmem(qCur2,SizeOf(qCur2^)*(nodes));

getmem(prevTranslation,SizeOf(prevTranslation^)*(nodes));


end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tactor.free;
begin
freemem(nodeParent,SizeOf(nodeParent^)*(nodes));
freemem(vertex,SizeOf(vertex^)*(vertexes));
{a} freemem(localized,SizeOf(localized^)*(vertexes));
freemem(triangle,SizeOf(triangle^)*(triangles));

freemem(rotation,SizeOf(rotation^)*(nodes));
freemem(translation,SizeOf(translation^)*(nodes));
{a} freemem(m_absolute,SizeOf(m_absolute^)*(nodes));
    freemem(m_relative,SizeOf(m_relative^)*(nodes));
{a} freemem(r_absolute,SizeOf(r_absolute^)*(nodes));
    freemem(r_relative,SizeOf(r_relative^)*(nodes));



freemem(qPrev,SizeOf(qPrev^)*(nodes));
freemem(qCur,SizeOf(qCur^)*(nodes));

{b} freemem(qCur1,SizeOf(qCur1^)*(nodes));
{b} freemem(qCur2,SizeOf(qCur2^)*(nodes));

freemem(prevTranslation,SizeOf(prevTranslation^)*(nodes));
end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tactor.localize;
var i:integer;
begin

for i:=0 to nodes-1 do begin
                       (r_relative+sizeOf(r_relative^)*i)^.setRotationRadians((rotation+sizeOf(rotation^)*i)^);
                       (r_relative+sizeOf(r_relative^)*i)^.setTranslation((translation+sizeOf(translation^)*i)^);
            if (nodeParent+sizeof(nodeParent^)*i)^<>255 then begin
                 (m_absolute+sizeOf(m_absolute^)*i)^:=(m_absolute+sizeOf(m_absolute^)* ((( (nodeParent+sizeof(nodeParent^)*i)^ ))) )^;
                 (m_absolute+sizeOf(m_absolute^)*i)^.postMultiply((r_relative+sizeOf(r_relative^)*i)^)
                                           end else
                                           begin
                                           (m_absolute+sizeOf(m_absolute^)*i)^:=(r_relative+sizeOf(r_relative^)*i)^
                                           end;
                     end;


for i:=0 to vertexes-1 do begin
(localized+sizeOf(localized^)*i)^:=(vertex+sizeOf(vertex^)*i)^;
(m_absolute+sizeOf(m_absolute^)*((vertex+sizeOf(vertex^)*i)^.parentId))^.inverseTranslateVect((localized+sizeOf(localized^)*i)^.coord);
(m_absolute+sizeOf(m_absolute^)*((vertex+sizeOf(vertex^)*i)^.parentId))^.inverseRotateVect((localized+sizeOf(localized^)*i)^.coord);


//{n} (m_absolute+sizeOf(m_absolute^)*((vertex+sizeOf(vertex^)*i)^.parentId))^.inverseTranslateVect((localized+sizeOf(localized^)*i)^.norm);
{n} (m_absolute+sizeOf(m_absolute^)*((vertex+sizeOf(vertex^)*i)^.parentId))^.inverseRotateVect((localized+sizeOf(localized^)*i)^.norm);

//(vertex+sizeOf(vertex^)*i)^:=(localized+sizeOf(localized^)*i)^;
end;


end;



// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tactor.upload(txpath,fname:string);
var f2:file;
    header:array[0..12]of char;
    i:integer;

begin
assign(f2,fname);
reset(f2);
blockread(f2,header,sizeof(header));
blockread(f2,vertexes,sizeof(vertexes));
blockread(f2,triangles,sizeof(triangles));
blockread(f2,texturesN,sizeof(texturesN));
blockread(f2,nodes,sizeof(nodes)); //count=nodes (for refskel - reference)

alloc;

blockread(f2,(vertex^),sizeof(vertex^)*vertexes);
blockread(f2,(triangle^),sizeof(triangle^)*triangles);
blockread(f2,(nodeParent^),sizeOf(nodeParent^)*nodes); //d
blockread(f2,(translation^),sizeOf(translation^)*nodes); //d
blockread(f2,(rotation^),sizeOf(rotation^)*nodes); //u

blockread(f2,texturelist,sizeof(texturelist));

close(f2);

for i:=0 to texturesN-1 do
 begin
  Textures^[i].Load({'ballon.txr'}txpath+StringWordGet (texturelist, '|', i+1));
  (material+sizeOf(material^)*i)^:=i;
 end;

localize;
end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tactor.render;
var i:integer;
    j:integer;
begin

           LockObj(zzz);

  for i:=0 to triangles-1 do
  begin
  if (triangle+sizeOf(triangle^)*i)^.textureId<>255 then
   with locpoly^[0] do
     st:=GetStatus(UtGuroudTex,BlendNone,0{TcMULRGB},
                   {!}(material+sizeOf(material^)*(triangle+sizeOf(triangle^)*i)^.textureId)^,
                   0,0);

//  TRIANGLE;
  for j:=0 to 2 do begin

            with locvert^[j].tcrd do
             begin
              u:=round( 1023 * (triangle+sizeOf(triangle^)*i)^.coord[j].u );
              v:=round( 1023 * (triangle+sizeOf(triangle^)*i)^.coord[j].v );
             end;


            with locvert^[j].nr do
             begin
              x:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[0];
              y:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[1];
              z:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[2];
             end;

            with locvert^[j].crd do
             begin
              x:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[0]*100;
              y:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[1]*100;
              z:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[2]*100;
             end;


  end;
  // triangleEND
  zzz.render;
  end;

(*

           LockObj(zzz);
           for i:=0 to 3 do
             with locvert^[i].crd do
             begin
               x:=x-100;
               z:=0;
             end;
              with locpoly^[0] do
     st:=GetStatus(UtGuroudTex,BlendNone,TcMULRGB,0,0,FCCULL);
     zzz.render;



           LockObj(zzz);
           for i:=0 to 3 do
             with locvert^[i].crd do
             begin
               x:=x+100;
               z:=4000;
             end;
             with locpoly^[0] do
           st:=GetStatus(UtGuroudTex,BlendNone,TcMULRGB,1,0,FCCULL);
     zzz.render;
  *)

end;


// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tactor.renderNT;
var i:integer;
    j:integer;
begin
  for i:=0 to triangles-1 do
  begin
//  if (triangle+sizeOf(triangle^)*i)^.textureId<>255 then
//  glBindTexture(GL_TEXTURE_2D,(material+sizeOf(material^)*(triangle+sizeOf(triangle^)*i)^.textureId)^);
//  TRIANGLE;
  for j:=0 to 2 do begin

            with locvert^[j].tcrd do
             begin
              u:=round( 1023 * (triangle+sizeOf(triangle^)*i)^.coord[j].u );
              v:=round( 1023 * (triangle+sizeOf(triangle^)*i)^.coord[j].v );
             end;


            with locvert^[j].nr do
             begin
              x:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[0];
              y:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[1];
              z:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[2];
             end;

            with locvert^[j].crd do
             begin
              x:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[0];
              y:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[1];
              z:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[2];
             end;


  end;
  // triangleEND
  zzz.render;
  end;

end;



// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tactor.animate(var ani:Tanimation;dir:integer);
var i,k:integer;
begin

//fps         : single;
//curtime     : single;
//curframe    : word;

  delta:=elapsed*ani.fps{/1000};

{catch errors}      if curframe>ani.frames-1 then curframe:=0;

  curtime:=curtime+delta;
  if curtime>1 then begin
     curtime:=0;

         for i:=0 to ani.nodes-1 do begin
         with ani do (qPrev+sizeOf(qPrev^)*i)^:=(qCur+sizeOf(qCur^)*i)^;
         (prevTranslation+sizeOf(prevTranslation^)*i)^:=(ani.translation+sizeOf(ani.translation^)*(i+curframe*ani.nodes))^;
         (qCur+sizeOf(qCur^)*i)^.fromangles((ani.rotation+sizeOf(ani.rotation^)*(i+curframe*ani.nodes))^);
         end;

      if curframe+dir>ani.frames-1 then curframe:=0
      else if curframe+dir<0 then curframe:=ani.frames-1 else
      curframe:=curframe+dir{1};



  end;




for i:=0 to ani.nodes-1 do begin

          for k:=0 to 2 do
          (translation+sizeOf(translation^)*i)^[k]:=(prevTranslation+sizeOf(prevTranslation^)*i)^[k]+
          ((ani.translation+sizeOf(ani.translation^)*(i+curframe*ani.nodes))^[k]-(prevTranslation+sizeOf(prevTranslation^)*i)^[k])*curTime;

//if dir>0 then

if ani.fix then
walk:=( (ani.translation+sizeOf(ani.translation^)*(0+2*ani.nodes))^[1] - (ani.translation+sizeOf(ani.translation^)*(0+1*ani.nodes))^[1] )*delta
else
if curframe<>ani.frames-1 then
if curframe<>0 then    walk:=((ani.translation+sizeOf(ani.translation^)*(0+curframe*ani.nodes))^[1]-(prevTranslation+sizeOf(prevTranslation^)*0)^[1])*delta;

{
if dir<0 then
if curframe<>ani.frames-1 then    walk:=((ani.translation+sizeOf(ani.translation^)*(0+curframe*ani.nodes))^[1]-(prevTranslation+sizeOf(prevTranslation^)*0)^[1])*delta;
}
          if ani.fix then (translation+sizeOf(translation^)*0)^[1]:=0;


     with ani do qFinal.slerp((qPrev+sizeOf(qPrev^)*i)^,(qCur+sizeOf(qCur^)*i)^,curTime);

                       (r_relative+sizeOf(r_relative^)*i)^.setRotationQuaternion(qFinal);
                       (r_relative+sizeOf(r_relative^)*i)^.setTranslation((translation+sizeOf(translation^)*i)^);

            if (nodeParent+sizeof(nodeParent^)*i)^<>255 then begin
                 (m_absolute+sizeOf(m_absolute^)*i)^:=(m_absolute+sizeOf(m_absolute^)* ((( (nodeParent+sizeof(nodeParent^)*i)^ ))) )^;
                 (m_absolute+sizeOf(m_absolute^)*i)^.postMultiply((r_relative+sizeOf(r_relative^)*i)^)
                                           end else
                                           begin
                                           (m_absolute+sizeOf(m_absolute^)*i)^:=(r_relative+sizeOf(r_relative^)*i)^
                                           end;
                     end;


for i:=0 to vertexes-1 do begin
(vertex+sizeOf(vertex^)*i)^:=(localized+sizeOf(localized^)*i)^;
    (vertex+sizeOf(vertex^)*i)^.coord:=transform((vertex+sizeOf(vertex^)*i)^.coord,(m_absolute+sizeOf(m_absolute^)*((vertex+sizeOf(vertex^)*i)^.parentId))^);
{n} (vertex+sizeOf(vertex^)*i)^.norm:=transform3((vertex+sizeOf(vertex^)*i)^.norm,(m_absolute+sizeOf(m_absolute^)*((vertex+sizeOf(vertex^)*i)^.parentId))^);
end;

end;






// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tactor.animateBlend(var ani,ani2:Tanimation;percent:single;dir:integer);
var i,k:integer;
begin

//fps         : single;
//curtime     : single;
//curframe    : word;

  delta:=elapsed*ani.fps{/1000};

  {catch errors}      if curframe>ani.frames-1 then curframe:=0;

  curtime:=curtime+delta;
  if curtime>1 then begin
     curtime:=0;

         for i:=0 to ani.nodes-1 do begin
         with ani do (qPrev+sizeOf(qPrev^)*i)^:=(qCur+sizeOf(qCur^)*i)^;
         (prevTranslation+sizeOf(prevTranslation^)*i)^:=(ani.translation+sizeOf(ani.translation^)*(i+curframe*ani.nodes))^;
         (qCur1+sizeOf(qCur1^)*i)^.fromangles((ani.rotation+sizeOf(ani.rotation^)*(i+curframe*ani.nodes))^);
         (qCur2+sizeOf(qCur2^)*i)^.fromangles((ani2.rotation+sizeOf(ani2.rotation^)*(i+curframe*ani2.nodes))^);
         (qCur+sizeOf(qCur^)*i)^.slerp((qCur1+sizeOf(qCur1^)*i)^,(qCur2+sizeOf(qCur2^)*i)^,percent);
         end;

      if curframe+dir>ani.frames-1 then curframe:=0
      else if curframe+dir<0 then curframe:=ani.frames-1 else
      curframe:=curframe+dir{1};



  end;




for i:=0 to ani.nodes-1 do begin

          for k:=0 to 2 do
          (translation+sizeOf(translation^)*i)^[k]:=(prevTranslation+sizeOf(prevTranslation^)*i)^[k]+
          ((ani.translation+sizeOf(ani.translation^)*(i+curframe*ani.nodes))^[k]-(prevTranslation+sizeOf(prevTranslation^)*i)^[k])*curTime;


if curframe<>0 then   walk:=((ani.translation+sizeOf(ani.translation^)*(0+curframe*ani.nodes))^[1]-(prevTranslation+sizeOf(prevTranslation^)*0)^[1])*delta;
          if ani.fix then (translation+sizeOf(translation^)*0)^[1]:=0;


     with ani do qFinal.slerp((qPrev+sizeOf(qPrev^)*i)^,(qCur+sizeOf(qCur^)*i)^,curTime);

                       (r_relative+sizeOf(r_relative^)*i)^.setRotationQuaternion(qFinal);
                       (r_relative+sizeOf(r_relative^)*i)^.setTranslation((translation+sizeOf(translation^)*i)^);

            if (nodeParent+sizeof(nodeParent^)*i)^<>255 then begin
                 (m_absolute+sizeOf(m_absolute^)*i)^:=(m_absolute+sizeOf(m_absolute^)* ((( (nodeParent+sizeof(nodeParent^)*i)^ ))) )^;
                 (m_absolute+sizeOf(m_absolute^)*i)^.postMultiply((r_relative+sizeOf(r_relative^)*i)^)
                                           end else
                                           begin
                                           (m_absolute+sizeOf(m_absolute^)*i)^:=(r_relative+sizeOf(r_relative^)*i)^
                                           end;
                     end;


for i:=0 to vertexes-1 do begin
(vertex+sizeOf(vertex^)*i)^:=(localized+sizeOf(localized^)*i)^;
(vertex+sizeOf(vertex^)*i)^.coord:=transform((vertex+sizeOf(vertex^)*i)^.coord,(m_absolute+sizeOf(m_absolute^)*((vertex+sizeOf(vertex^)*i)^.parentId))^);
{n} (vertex+sizeOf(vertex^)*i)^.norm:=transform3((vertex+sizeOf(vertex^)*i)^.norm,(m_absolute+sizeOf(m_absolute^)*((vertex+sizeOf(vertex^)*i)^.parentId))^);
end;

end;










// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tanimation.alloc;
begin
getmem(rotation,SizeOf(rotation^)*(count));
getmem(translation,SizeOf(translation^)*(count));

end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tanimation.free;
begin
freemem(rotation,SizeOf(rotation^)*(count));
freemem(translation,SizeOf(translation^)*(count));

end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tanimation.upload(fname:string);
var f2:file;
var header:array[0..12]of char;
begin
assign(f2,fname);
reset(f2);
blockread(f2,header,sizeof(header));
blockread(f2,nodes,sizeof(nodes)); //count=nodes (for refskel - reference)
blockread(f2,frames,sizeof(frames));

{new}count:=frames*nodes;

alloc;

blockread(f2,(translation^),sizeOf(translation^)*count); //d
blockread(f2,(rotation^),sizeOf(rotation^)*count); //u


close(f2);
end;


// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tobject.alloc;
begin

//triangle primitive for Polyx 3d engine
     zzz.init;
     zzz.create(3,1); // 3 vertexes = 1 face
     lockobj(zzz);

     setvert(0,
             0,0,0,
             255,255,255,255,
             0,0);
     setvert(1,
             1000,0,0,
             255,255,255,255,
             1023,0);
     setvert(2,
             1000,1000,0,
             255,255,255,255,
             1023,1023);

     setface(0,
             0,1,2,-1,
             GetStatus(UtGuroudTex,BlendNone,TcMULRGB,0,0,0),255);


getmem(vertex,SizeOf(vertex^)*(vertexes));
{a}getmem(localized,SizeOf(localized^)*(vertexes));
getmem(triangle,SizeOf(triangle^)*(triangles));

getmem(material,SizeOf(material^)*(texturesN));
end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tobject.free;
begin
freemem(vertex,SizeOf(vertex^)*(vertexes));
{a}freemem(localized,SizeOf(localized^)*(vertexes));
freemem(triangle,SizeOf(triangle^)*(triangles));

freemem(material,SizeOf(material^)*(texturesN));
end;


// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tobject.localize;
var i:integer;
begin
for i:=0 to vertexes-1 do
(localized+sizeOf(localized^)*i)^:=(vertex+sizeOf(vertex^)*i)^;
end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tobject.upload(txpath,fname:string);
var f2:file;
    header:array[0..12]of char;
    i:integer;

begin
assign(f2,fname);
reset(f2);
blockread(f2,header,sizeof(header));
blockread(f2,vertexes,sizeof(vertexes));
blockread(f2,triangles,sizeof(triangles));
blockread(f2,texturesN,sizeof(texturesN));

alloc;

blockread(f2,(vertex^),sizeof(vertex^)*vertexes);
blockread(f2,(triangle^),sizeof(triangle^)*triangles);

blockread(f2,texturelist,sizeof(texturelist));

close(f2);

(*
  for i:=0 to texturesN-1 do
  LoadTextureJPG(txpath+StringWordGet (texturelist, '|', i+1),(material+sizeOf(material^)*i)^);
*)

for i:=0 to texturesN-1 do
 begin
  Textures^[i].Load({'ballon.txr'}txpath+StringWordGet (texturelist, '|', i+1));
  (material+sizeOf(material^)*i)^:=i;
 end;


localize;
end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tobject.render;
var i:integer;
    j:integer;
begin
           LockObj(zzz);

  for i:=0 to triangles-1 do
  begin
  if (triangle+sizeOf(triangle^)*i)^.textureId<>255 then
   with locpoly^[0] do
     st:=GetStatus(UtGuroudTex,BlendNone,TcMULRGB,
                   {!}(material+sizeOf(material^)*(triangle+sizeOf(triangle^)*i)^.textureId)^,
                   0,0);

//  TRIANGLE;
  for j:=0 to 2 do begin

            with locvert^[j].tcrd do
             begin
              u:=round( 1023 * (triangle+sizeOf(triangle^)*i)^.coord[j].u );
              v:=round( 1023 * (triangle+sizeOf(triangle^)*i)^.coord[j].v );
             end;


            with locvert^[j].nr do
             begin
              x:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[0];
              y:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[1];
              z:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[2];
             end;

            with locvert^[j].crd do
             begin
              x:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[0];
              y:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[1];
              z:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[2];
             end;



  end;
  // triangleEND
  zzz.render;

  end;
 end;

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tobject.renderNT;
var i:integer;
    j:integer;
begin
  for i:=0 to triangles-1 do
  begin
//  if (triangle+sizeOf(triangle^)*i)^.textureId<>255 then
//  glBindTexture(GL_TEXTURE_2D,(material+sizeOf(material^)*(triangle+sizeOf(triangle^)*i)^.textureId)^);
//  TRIANGLES;
  for j:=0 to 2 do begin

            with locvert^[j].tcrd do
             begin
              u:=round( 1023 * (triangle+sizeOf(triangle^)*i)^.coord[j].u );
              v:=round( 1023 * (triangle+sizeOf(triangle^)*i)^.coord[j].v );
             end;


            with locvert^[j].nr do
             begin
              x:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[0];
              y:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[1];
              z:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.norm[2];
             end;

            with locvert^[j].crd do
             begin
              x:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[0];
              y:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[1];
              z:=(vertex+sizeOf(vertex^)*( (triangle+sizeOf(triangle^)*i)^.vertexId[j]))^.coord[2];
             end;

           zzz.render;

  end;
  // trianglesEND
  end;
 end;


// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
procedure Tobject.attach(act:Tactor;node:byte);
var i:integer;
begin
     for i:=0 to vertexes-1 do begin
     (vertex+sizeOf(vertex^)*i)^:=(localized+sizeOf(localized^)*i)^;
     (vertex+sizeOf(vertex^)*i)^.coord:=transform((vertex+sizeOf(vertex^)*i)^.coord,(act.m_absolute+sizeOf(act.m_absolute^)*node)^);
 {n} (vertex+sizeOf(vertex^)*i)^.norm:=transform3((vertex+sizeOf(vertex^)*i)^.norm,  (act.m_absolute+sizeOf(act.m_absolute^)*node)^);
     end;
end;

end.