// translated to Pascal by Georgy Moshkin (tmtlib@narod.ru)

unit _math;

interface

type TVector4s = array[0..3] of single;
type TVector3s = array[0..2] of single;
type TVector2s = array[0..1] of single;

type TVector3w = array[0..2] of word;



type Quaternion = object {Quaternion object}
                  m_quat: array[0..3] of single; {Variables}
                  Procedure fromAngles(angles:TVector3s); { Procedures }
                  Procedure slerp(q1,q2:quaternion;interp:single);
                  Procedure inverse;
                  end;

type Matrix = object {Matirx object}
              m_matrix:array[0..15] of single; {Variables}
              Procedure PostMultiply(_m: Matrix); {Procedures}
              Procedure SetTranslation(trans:TVector3s);
              Procedure SetRotationRadians(angles:TVector3s);
              Procedure InverseRotateVect(var pVect:TVector3s);
              Procedure InverseTranslateVect(var pVect:TVector3s{4});
              Procedure setRotationQuaternion(quat:Quaternion);
              end;


function transform(this_vector:TVector3s{4};ourm_matrix:matrix):TVector3s{4};
function transform3(this_vector:TVector3s;ourm_matrix:matrix):TVector3s;
implementation

uses math;

{Vector object Procedures}


{Quaternion object Procedures}

////////////////////////////////////////////////////////////////////////////////
Procedure Quaternion.fromAngles(angles:TVector3s{4});

var angle             : single;
    sr,sp,sy,cr,cp,cy : single;
    crcp,srsp         : single;

begin
     angle:=angles[2]*0.5;
     sy:=sin( angle );
     cy:=cos( angle );
     angle:= angles[1]*0.5;
     sp:= sin( angle );
     cp:= cos( angle );
     angle:= angles[0]*0.5;
     sr:= sin( angle );
     cr:= cos( angle );

     crcp:= cr*cp;
     srsp:= sr*sp;

     m_quat[0]:=sr*cp*cy-cr*sp*sy;
     m_quat[1]:=cr*sp*cy+sr*cp*sy;
     m_quat[2]:=crcp*sy-srsp*cy;
     m_quat[3]:=crcp*cy+srsp*sy;
end;


////////////////////////////////////////////////////////////////////////////////
Procedure Quaternion.slerp( q1,q2:quaternion; interp:single);

var i      : integer;
    a,b    : single;

    cosom       : single;
    sclq1,sclq2 : single;
    omega,sinom : single;

begin
     a:=0;
     b:=0;

     for i:=0 to 3 do begin
                           a+:=( q1.m_quat[i]-q2.m_quat[i] )*( q1.m_quat[i]-q2.m_quat[i] );
                           b+:=( q1.m_quat[i]+q2.m_quat[i] )*( q1.m_quat[i]+q2.m_quat[i] );
                      end;

     if ( a > b ) then q2.inverse;

     cosom:=q1.m_quat[0]*q2.m_quat[0]+q1.m_quat[1]*q2.m_quat[1]
           +q1.m_quat[2]*q2.m_quat[2]+q1.m_quat[3]*q2.m_quat[3];

     if (( 1.0+cosom ) > 0.00000001 ) then begin{#}
     if (( 1.0-cosom ) > 0.00000001 ) then begin
                                                omega:= arccos( cosom );
                                                sinom:= sin( omega );
                                                sclq1:= sin(( 1.0-interp )*omega )/sinom;
                                                sclq2:= sin( interp*omega )/sinom;
                                           end else
                                                   begin
                                                        sclq1:= 1.0-interp;
                                                        sclq2:= interp;
                                                   end;
                                           for i:=0 to 3 do
                                           m_quat[i]:=sclq1*q1.m_quat[i]
                                                     +sclq2*q2.m_quat[i];
                                           end{#} else
                                                      begin
                                                           m_quat[0]:=-q1.m_quat[1];
                                                           m_quat[1]:=q1.m_quat[0];
                                                           m_quat[2]:=-q1.m_quat[3];
                                                           m_quat[3]:=q1.m_quat[2];

                                                           sclq1:= sin(( 1.0-interp )*0.5*PI );
                                                           sclq2:= sin( interp*0.5*PI );

                                                           for i:=0 to 2 do
                                                           m_quat[i]:=sclq1*q1.m_quat[i]
                                                                     +sclq2*m_quat[i];
                                                      end;

                                                  end;



////////////////////////////////////////////////////////////////////////////////
Procedure Quaternion.inverse;
begin
     m_quat[0]:= -m_quat[0];
     m_quat[1]:= -m_quat[1];
     m_quat[2]:= -m_quat[2];
     m_quat[3]:= -m_quat[3];
end;


{Matrix object Procedures}

////////////////////////////////////////////////////////////////////////////////
procedure Matrix.PostMultiply(_m:Matrix);
var t_m_matrix:Matrix;
begin
     t_m_matrix.m_matrix[0]:= m_matrix[0]*_m.m_matrix[0] + m_matrix[4]*_m.m_matrix[1] + m_matrix[8]*_m.m_matrix[2];
     t_m_matrix.m_matrix[1]:= m_matrix[1]*_m.m_matrix[0] + m_matrix[5]*_m.m_matrix[1] + m_matrix[9]*_m.m_matrix[2];
     t_m_matrix.m_matrix[2]:= m_matrix[2]*_m.m_matrix[0] + m_matrix[6]*_m.m_matrix[1] + m_matrix[10]*_m.m_matrix[2];
     t_m_matrix.m_matrix[3]:= 0;

     t_m_matrix.m_matrix[4]:= m_matrix[0]*_m.m_matrix[4] + m_matrix[4]*_m.m_matrix[5] + m_matrix[8]*_m.m_matrix[6];
     t_m_matrix.m_matrix[5]:= m_matrix[1]*_m.m_matrix[4] + m_matrix[5]*_m.m_matrix[5] + m_matrix[9]*_m.m_matrix[6];
     t_m_matrix.m_matrix[6]:= m_matrix[2]*_m.m_matrix[4] + m_matrix[6]*_m.m_matrix[5] + m_matrix[10]*_m.m_matrix[6];
     t_m_matrix.m_matrix[7]:= 0;

     t_m_matrix.m_matrix[8]:= m_matrix[0]*_m.m_matrix[8] + m_matrix[4]*_m.m_matrix[9] + m_matrix[8]*_m.m_matrix[10];
     t_m_matrix.m_matrix[9]:= m_matrix[1]*_m.m_matrix[8] + m_matrix[5]*_m.m_matrix[9] + m_matrix[9]*_m.m_matrix[10];
     t_m_matrix.m_matrix[10]:= m_matrix[2]*_m.m_matrix[8] + m_matrix[6]*_m.m_matrix[9] + m_matrix[10]*_m.m_matrix[10];
     t_m_matrix.m_matrix[11]:= 0;

     t_m_matrix.m_matrix[12]:= m_matrix[0]*_m.m_matrix[12] + m_matrix[4]*_m.m_matrix[13] + m_matrix[8]*_m.m_matrix[14] + m_matrix[12];
     t_m_matrix.m_matrix[13]:= m_matrix[1]*_m.m_matrix[12] + m_matrix[5]*_m.m_matrix[13] + m_matrix[9]*_m.m_matrix[14] + m_matrix[13];
     t_m_matrix.m_matrix[14]:= m_matrix[2]*_m.m_matrix[12] + m_matrix[6]*_m.m_matrix[13] + m_matrix[10]*_m.m_matrix[14] + m_matrix[14];
     t_m_matrix.m_matrix[15]:= 1;

     m_matrix:=t_m_matrix.m_matrix;
end;


////////////////////////////////////////////////////////////////////////////////
Procedure Matrix.SetTranslation(trans:TVector3s{4});
begin
     m_matrix[12]:= trans[0];
     m_matrix[13]:= trans[1];
     m_matrix[14]:= trans[2];
end;

////////////////////////////////////////////////////////////////////////////////
Procedure Matrix.SetRotationRadians(angles:TVector3s{4});
var cr,sr,cp,sp,cy,sy,
    srsp,crsp         : single;
begin
     cr:= cos(angles[0]);
     sr:= sin(angles[0]);
     cp:= cos(angles[1]);
     sp:= sin(angles[1]);
     cy:= cos(angles[2]);
     sy:= sin(angles[2]);

     m_matrix[0]:= cp*cy;
     m_matrix[1]:= cp*sy;
     m_matrix[2]:= -sp;

     srsp:= sr*sp;
     crsp:= cr*sp;

     m_matrix[4]:= srsp*cy-cr*sy;
     m_matrix[5]:= srsp*sy+cr*cy;
     m_matrix[6]:= sr*cp;

     m_matrix[8]:= crsp*cy+sr*sy;
     m_matrix[9]:= crsp*sy-sr*cy;
     m_matrix[10]:= cr*cp;
end;


////////////////////////////////////////////////////////////////////////////////
Procedure Matrix.InverseRotateVect(var pVect:TVector3s{4});
var temp : TVector3s{4};
begin
     temp:=pVect;

     temp[0]:= pVect[0]*m_matrix[0]+pVect[1]*m_matrix[1]+pVect[2]*m_matrix[2];
     temp[1]:= pVect[0]*m_matrix[4]+pVect[1]*m_matrix[5]+pVect[2]*m_matrix[6];
     temp[2]:= pVect[0]*m_matrix[8]+pVect[1]*m_matrix[9]+pVect[2]*m_matrix[10];

     pVect:=temp;
end;

////////////////////////////////////////////////////////////////////////////////
Procedure Matrix.InverseTranslateVect(var pVect:TVector3s{4});
var temp : TVector3s{4};
begin
     temp:=pVect;

     temp[0]:= pVect[0]-m_matrix[12];
     temp[1]:= pVect[1]-m_matrix[13];
     temp[2]:= pVect[2]-m_matrix[14];

     pVect:=temp;
end;

////////////////////////////////////////////////////////////////////////////////
function transform(this_vector:TVector3s{4};ourm_matrix:matrix):TVector3s{4};
var our_vector:TVector3s{4};
begin
     our_vector[0]:= this_vector[0]*ourm_matrix.m_matrix[0]+this_vector[1]*ourm_matrix.m_matrix[4]+this_vector[2]*ourm_matrix.m_matrix[8]+ourm_matrix.m_matrix[12];
     our_vector[1]:= this_vector[0]*ourm_matrix.m_matrix[1]+this_vector[1]*ourm_matrix.m_matrix[5]+this_vector[2]*ourm_matrix.m_matrix[9]+ourm_matrix.m_matrix[13];
     our_vector[2]:= this_vector[0]*ourm_matrix.m_matrix[2]+this_vector[1]*ourm_matrix.m_matrix[6]+this_vector[2]*ourm_matrix.m_matrix[10]+ourm_matrix.m_matrix[14];
//     our_vector[3]:= this_vector[0]*ourm_matrix.m_matrix[3]+this_vector[1]*ourm_matrix.m_matrix[7]+this_vector[2]*ourm_matrix.m_matrix[11]+ourm_matrix.m_matrix[15];

     transform:=our_vector;
end;


function transform3(this_vector:TVector3s;ourm_matrix:matrix):TVector3s;
var our_vector:TVector3s{4};
begin
     our_vector[0]:= this_vector[0]*ourm_matrix.m_matrix[0]+this_vector[1]*ourm_matrix.m_matrix[4]+this_vector[2]*ourm_matrix.m_matrix[8];
     our_vector[1]:= this_vector[0]*ourm_matrix.m_matrix[1]+this_vector[1]*ourm_matrix.m_matrix[5]+this_vector[2]*ourm_matrix.m_matrix[9];
     our_vector[2]:= this_vector[0]*ourm_matrix.m_matrix[2]+this_vector[1]*ourm_matrix.m_matrix[6]+this_vector[2]*ourm_matrix.m_matrix[10];
     transform3:=our_vector;
end;



////////////////////////////////////////////////////////////////////////////////
Procedure Matrix.setRotationQuaternion(quat:Quaternion);
begin
     m_matrix[0]:= ( 1.0 - 2.0*quat.m_quat[1]*quat.m_quat[1] - 2.0*quat.m_quat[2]*quat.m_quat[2]);
     m_matrix[1]:= ( 2.0*quat.m_quat[0]*quat.m_quat[1] + 2.0*quat.m_quat[3]*quat.m_quat[2] );
     m_matrix[2]:= ( 2.0*quat.m_quat[0]*quat.m_quat[2] - 2.0*quat.m_quat[3]*quat.m_quat[1] );

     m_matrix[4]:= ( 2.0*quat.m_quat[0]*quat.m_quat[1] - 2.0*quat.m_quat[3]*quat.m_quat[2] );
     m_matrix[5]:= ( 1.0 - 2.0*quat.m_quat[0]*quat.m_quat[0] - 2.0*quat.m_quat[2]*quat.m_quat[2] );
     m_matrix[6]:= ( 2.0*quat.m_quat[1]*quat.m_quat[2] + 2.0*quat.m_quat[3]*quat.m_quat[0] );

     m_matrix[8]:= ( 2.0*quat.m_quat[0]*quat.m_quat[2] + 2.0*quat.m_quat[3]*quat.m_quat[1] );
     m_matrix[9]:= ( 2.0*quat.m_quat[1]*quat.m_quat[2] - 2.0*quat.m_quat[3]*quat.m_quat[0] );
     m_matrix[10]:= ( 1.0 - 2.0*quat.m_quat[0]*quat.m_quat[0] - 2.0*quat.m_quat[1]*quat.m_quat[1] );
end;


////////////////////////////////////////////////////////////////////////////////



end.